@@ -9,14 +9,20 @@
 #define GIOP_CHUNK_ALIGN 8
 #define GIOP_CHUNK_SIZE (GIOP_CHUNK_ALIGN * 256)
 
+static gboolean giop_blank_wire_data = FALSE;
 static GSList *send_buffer_list = NULL;
 static GMutex *send_buffer_list_lock = NULL;
 
 static const char giop_zero_buf [GIOP_CHUNK_ALIGN * 10] = {0};
 
 void
-giop_send_buffer_init (void)
+giop_send_buffer_init (gboolean wipe)
 {
+#ifdef ORBIT_PURIFY
+	giop_blank_wire_data = TRUE;
+#else
+	giop_blank_wire_data = wipe;
+#endif
 	send_buffer_list_lock = linc_mutex_new ();
 }
 
@@ -45,51 +51,54 @@ giop_send_buffer_use_request (GIOPVersion giop_version,
 			      const struct iovec *operation_vec,
 			      const struct iovec *principal_vec)
 {
-  GIOPSendBuffer *buf = giop_send_buffer_use(giop_version);
-  struct iovec zerovec;
+	GIOPSendBuffer *buf = giop_send_buffer_use (giop_version);
+	struct iovec zerovec;
 
-  if(!principal_vec)
-    {
-      zerovec.iov_base = (gpointer) giop_zero_buf;
-      zerovec.iov_len = sizeof(CORBA_unsigned_long);
-      principal_vec = &zerovec;
-    }
+	if(!principal_vec) {
+		zerovec.iov_base = (gpointer) giop_zero_buf;
+		zerovec.iov_len = sizeof (CORBA_unsigned_long);
+		principal_vec = &zerovec;
+	}
 
-  buf->msg.header.message_type = GIOP_REQUEST;
-  giop_send_buffer_align(buf, sizeof(CORBA_unsigned_long));
-  buf->msg.u.request_1_0.request_id = request_id;
-  buf->msg.u.request_1_0.response_expected = response_expected;
+	buf->msg.header.message_type = GIOP_REQUEST;
+	giop_send_buffer_align (buf, sizeof(CORBA_unsigned_long));
 
-  switch(giop_version)
-    {
-    case GIOP_1_0:
-    case GIOP_1_1:
-      giop_send_buffer_append(buf, (const guchar *)iop_service_context_data, sizeof(iop_service_context_data));
-      giop_send_buffer_append(buf, &buf->msg.u.request_1_0.request_id, sizeof(CORBA_unsigned_long));
-      giop_send_buffer_append(buf, &buf->msg.u.request_1_0.response_expected, sizeof(CORBA_boolean));
-      giop_send_buffer_append_aligned(buf, &objkey->_length, sizeof(CORBA_unsigned_long));
-      giop_send_buffer_append(buf, objkey->_buffer, objkey->_length);
-      giop_send_buffer_align(buf, sizeof(CORBA_unsigned_long));
-      giop_send_buffer_append(buf, operation_vec->iov_base, operation_vec->iov_len);
-      giop_send_buffer_append(buf, principal_vec->iov_base, principal_vec->iov_len);
-      break;
-    case GIOP_1_2:
-      giop_send_buffer_align(buf, sizeof(CORBA_unsigned_long));
-      giop_send_buffer_append(buf, &buf->msg.u.request_1_0.request_id, sizeof(CORBA_unsigned_long));
-      giop_send_buffer_append(buf, &buf->msg.u.request_1_0.response_expected, sizeof(CORBA_boolean));
-      giop_send_buffer_append(buf, giop_zero_buf, 3);
-      giop_send_buffer_append(buf, &giop_1_2_target_type, 2); /* We always use GIOP::KeyAddr addressing - the only sane way */
-      giop_send_buffer_append_aligned(buf, &objkey->_length, sizeof(CORBA_unsigned_long));
-      giop_send_buffer_append(buf, objkey->_buffer, objkey->_length);
-      giop_send_buffer_align(buf, sizeof(CORBA_unsigned_long));
-      giop_send_buffer_append(buf, operation_vec->iov_base, operation_vec->iov_len);
-      giop_send_buffer_append(buf, (const guchar *)iop_service_context_data, sizeof(iop_service_context_data));
-      giop_send_buffer_align(buf, 8); /* alignment for the body */
-    default:
-      break;
-    }
+	switch (giop_version) {
+	case GIOP_1_0:
+	case GIOP_1_1:
+		buf->msg.u.request_1_0.request_id = request_id;
+		buf->msg.u.request_1_0.response_expected = response_expected;
 
-  return buf;
+		giop_send_buffer_append (buf, (const guchar *)iop_service_context_data, sizeof(iop_service_context_data));
+		giop_send_buffer_append (buf, &buf->msg.u.request_1_0.request_id, sizeof(CORBA_unsigned_long));
+		giop_send_buffer_append (buf, &buf->msg.u.request_1_0.response_expected, sizeof(CORBA_boolean));
+		giop_send_buffer_append_aligned (buf, &objkey->_length, sizeof(CORBA_unsigned_long));
+		giop_send_buffer_append (buf, objkey->_buffer, objkey->_length);
+		giop_send_buffer_align (buf, sizeof(CORBA_unsigned_long));
+		giop_send_buffer_append (buf, operation_vec->iov_base, operation_vec->iov_len);
+		giop_send_buffer_append (buf, principal_vec->iov_base, principal_vec->iov_len);
+		break;
+
+	case GIOP_1_2:
+		buf->msg.u.request_1_2.request_id = request_id;
+		buf->msg.u.request_1_2.response_flags = response_expected ? 0x3 /* SYNC_WITH_TARGET */ : 0x0 /* SYNC_NONE */;
+
+		giop_send_buffer_align (buf, sizeof(CORBA_unsigned_long));
+		giop_send_buffer_append (buf, &buf->msg.u.request_1_2.request_id, sizeof(CORBA_unsigned_long));
+		giop_send_buffer_append (buf, &buf->msg.u.request_1_2.response_flags, sizeof(CORBA_octet));
+		giop_send_buffer_append (buf, giop_zero_buf, 3);
+		giop_send_buffer_append (buf, &giop_1_2_target_type, 2); /* We always use GIOP::KeyAddr addressing - the only sane way */
+		giop_send_buffer_append_aligned (buf, &objkey->_length, sizeof(CORBA_unsigned_long));
+		giop_send_buffer_append (buf, objkey->_buffer, objkey->_length);
+		giop_send_buffer_align (buf, sizeof(CORBA_unsigned_long));
+		giop_send_buffer_append (buf, operation_vec->iov_base, operation_vec->iov_len);
+		giop_send_buffer_append (buf, (const guchar *)iop_service_context_data, sizeof(iop_service_context_data));
+		giop_send_buffer_align (buf, 8); /* alignment for the body */
+	default:
+		break;
+	}
+
+	return buf;
 }
 
 GIOPSendBuffer *
@@ -235,7 +244,7 @@ giop_send_buffer_append_real (GIOPSendBuffer *buf,
 
 /*
  * get_next_indirect:
- * @buf: the recv buffer with an exhausted indirect.
+ * @buf: the send buffer with an exhausted indirect.
  * @for_size_hint: for very large buffers specify this
  * so we don't allocate too much. If this is non 0 then
  * buf->indirect will contain at least this much space.
@@ -264,7 +273,18 @@ get_next_indirect (GIOPSendBuffer *buf, gulong for_size_hint)
 			new_size = GIOP_CHUNK_SIZE;
 
 		buf->indirects [max].size = new_size;
-		buf->indirects [max].ptr = g_malloc (new_size);
+
+		if (giop_blank_wire_data)
+			buf->indirects [max].ptr = g_malloc0 (new_size);
+		else
+			buf->indirects [max].ptr = g_malloc (new_size);
+
+		/*
+		 *   We assume that this is 8 byte aligned, for efficiency -
+		 * so we can align to the memory address rather than the offset
+		 * into the buffer.
+		 */
+		g_assert (((gulong)buf->indirects [max].ptr & 0x3) == 0);
 	}
 
 	buf->indirect = buf->indirects [max].ptr;
@@ -398,10 +418,8 @@ giop_send_buffer_write (GIOPSendBuffer *buf,
 	if (!non_block)
 		non_block = linc_write_options_new (FALSE);
 
-	if (buf->giop_version >= GIOP_1_2)
-		giop_send_buffer_align (buf, 8); /* Do tail align */
-
-	LINC_MUTEX_LOCK (cnx->outgoing_mutex);
+	/* FIXME: if a FRAGMENT, assert the 8 byte tail align,
+	   &&|| giop_send_buffer_align (buf, 8); */
 
 	retval = linc_connection_writev (
 		(LINCConnection *) cnx, buf->iovecs,
@@ -413,8 +431,6 @@ giop_send_buffer_write (GIOPSendBuffer *buf,
 
 	/* FIXME: we need to flag the connection disconnected on fatal error */
 
-	LINC_MUTEX_UNLOCK (cnx->outgoing_mutex);
-
 	return retval;
 }
 
@@ -422,7 +438,11 @@ giop_send_buffer_write (GIOPSendBuffer *buf,
 GIOPSendBuffer *
 giop_send_buffer_use (GIOPVersion giop_version)
 {
-	GIOPSendBuffer *retval;
+	GIOPSendBuffer *buf;
+
+	g_return_val_if_fail (
+		((int) giop_version) >= 0 &&
+		giop_version < GIOP_NUM_VERSIONS, NULL);
 
 	LINC_MUTEX_LOCK (send_buffer_list_lock);
 	if (send_buffer_list) {
@@ -434,33 +454,43 @@ giop_send_buffer_use (GIOPVersion giop_version)
 
 		LINC_MUTEX_UNLOCK (send_buffer_list_lock);
 
-		retval = ltmp->data;
+		buf = ltmp->data;
 		g_slist_free_1 (ltmp);
-		retval->num_used = retval->indirect_left = 0;
-		retval->num_indirects_used = 0;
+		buf->num_used = buf->indirect_left = 0;
+
+		if (giop_blank_wire_data) {
+			int i;
+
+			for (i = 0; i < buf->num_indirects_used; i++)
+				memset (buf->indirects [i].ptr, 0,
+					buf->indirects [i].size);
+		}
+
+		buf->num_indirects_used = 0;
 	} else {
 		LINC_MUTEX_UNLOCK (send_buffer_list_lock);
 
-		retval = g_new0 (GIOPSendBuffer, 1);
+		buf = g_new0 (GIOPSendBuffer, 1);
 
-		memcpy (retval->msg.header.magic, "GIOP", 4);
-		retval->msg.header.flags = GIOP_FLAG_ENDIANNESS;
-		retval->num_alloced = 8;
-		retval->iovecs = g_new (struct iovec, 8);
+		memcpy (buf->msg.header.magic, "GIOP", 4);
+		buf->msg.header.flags = GIOP_FLAG_ENDIANNESS;
+		buf->num_alloced = 8;
+		buf->iovecs = g_new (struct iovec, 8);
 	}
 
-	memcpy (retval->msg.header.version,
+
+	memcpy (buf->msg.header.version,
 		giop_version_ids [giop_version], 2);
-	retval->giop_version = giop_version;
+	buf->giop_version = giop_version;
 
-	g_assert (sizeof (retval->msg.header) == 12);
+	g_assert (sizeof (buf->msg.header) == 12);
 	giop_send_buffer_append_real (
-		retval, (guchar *)&retval->msg.header, 12);
+		buf, (guchar *)&buf->msg.header, 12);
 
-	retval->msg.header.message_size = 0;
-	retval->header_size = 12;
+	buf->msg.header.message_size = 0;
+	buf->header_size = 12;
 
-	return retval;
+	return buf;
 }
 
 void
