@@ -26,17 +26,18 @@ OrbitDebugFlags _orbit_debug_flags = ORBIT_DEBUG_NONE;
 /*
  * Command line option handling.
  */
-static gboolean     orbit_use_ipv4          = FALSE;
-static gboolean     orbit_use_ipv6          = FALSE; 
-static gboolean     orbit_use_usocks        = TRUE;
-static gboolean     orbit_use_irda          = FALSE;
-static gboolean     orbit_use_ssl           = FALSE;
-static gboolean     orbit_use_genuid_simple = FALSE;
-static gboolean     orbit_local_only       = FALSE;
-static gboolean     orbit_use_http_iors     = FALSE;
-static char        *orbit_ipsock            = NULL;
-static char        *orbit_ipname            = NULL;
-static char        *orbit_debug_options     = NULL;
+static gboolean     orbit_use_ipv4           = FALSE;
+static gboolean     orbit_use_ipv6           = FALSE; 
+static gboolean     orbit_use_usocks         = TRUE;
+static gint         orbit_initial_recv_limit = -1;
+static gboolean     orbit_use_irda           = FALSE;
+static gboolean     orbit_use_ssl            = FALSE;
+static gboolean     orbit_use_genuid_simple  = FALSE;
+static gboolean     orbit_local_only         = FALSE;
+static gboolean     orbit_use_http_iors      = FALSE;
+static char        *orbit_ipsock             = NULL;
+static char        *orbit_ipname             = NULL;
+static char        *orbit_debug_options      = NULL;
 
 void
 ORBit_ORB_start_servers (CORBA_ORB orb)
@@ -95,16 +96,29 @@ ORBit_ORB_start_servers (CORBA_ORB orb)
 	LINC_MUTEX_UNLOCK (orb->lock);
 }
 
+static void
+strip_object_profiles (gpointer o, gpointer b, gpointer c)
+{
+	CORBA_Object obj = o;
+	IOP_delete_profiles (obj->orb, &obj->profile_list);
+	IOP_delete_profiles (obj->orb, &obj->forward_locations);
+}
+
 static void
 ORBit_ORB_shutdown_servers (CORBA_ORB orb)
 {
 	LINC_MUTEX_LOCK (orb->lock);
 
+	if (orb->objrefs) {
+		g_hash_table_foreach (orb->objrefs,
+				      strip_object_profiles, NULL);
+		g_hash_table_destroy (orb->objrefs);
+		orb->objrefs = NULL;
+	}
+
 	IOP_shutdown_profiles (orb->profiles);
 	orb->profiles = NULL;
 
-	if (giop_threaded ())
-		g_warning ("FIXME: this needs delaying till worker thread exit");
 	g_slist_foreach (orb->servers, (GFunc) g_object_unref, NULL);
 	g_slist_free (orb->servers); 
 	orb->servers = NULL;
@@ -267,10 +281,13 @@ CORBA_ORB_init (int *argc, char **argv,
 #endif /* G_ENABLE_DEBUG */
 
 	genuid_init ();
-	
+
+	giop_recv_set_limit (orbit_initial_recv_limit);
 	giop_init (threaded,
 		   orbit_use_ipv4 || orbit_use_ipv6 ||
 		   orbit_use_irda || orbit_use_ssl);
+	_ORBit_object_init ();
+	ORBit_poa_init ();
 
 	ORBit_locks_initialize ();
 
@@ -946,7 +963,7 @@ void
 CORBA_ORB_run (CORBA_ORB          orb,
 	       CORBA_Environment *ev)
 {
-	linc_main_loop_run ();
+	giop_main_run ();
 }
 
 void
@@ -956,20 +973,22 @@ CORBA_ORB_shutdown (CORBA_ORB           orb,
 {
 	PortableServer_POA root_poa;
 
-	if (init_level > 0)
-		return;
-
 	root_poa = g_ptr_array_index (orb->adaptors, 0);
 	if (root_poa) {
 		PortableServer_POA_destroy (
 			root_poa, TRUE, wait_for_completion, ev);
-		if (ev->_major)
-			return;
+		if (ev->_major) {
+			if (wait_for_completion)
+				g_warning ("FIXME: wait for "
+					   "completion unimplemented");
+			else
+				return;
+		}
 	}
 
-	ORBit_ORB_shutdown_servers (orb);
-
 	giop_shutdown ();
+
+	ORBit_ORB_shutdown_servers (orb);
 }
 
 void
@@ -986,10 +1005,11 @@ CORBA_ORB_destroy (CORBA_ORB          orb,
 	if (init_level > 0)
 		return;
 
+	CORBA_ORB_shutdown (orb, TRUE, ev);
+
 	g_assert (_ORBit_orb == orb);
 	_ORBit_orb = NULL;
 
-	CORBA_ORB_shutdown (orb, TRUE, ev);
 	if (ev->_major)
 		return;
 
@@ -1131,25 +1151,26 @@ ORBit_proto_use (const char *name)
 }
 
 const ORBit_option orbit_supported_options[] = {
-	{ "ORBid",           ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
-	{ "ORBImplRepoIOR",  ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
-	{ "ORBIfaceRepoIOR", ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
-	{ "ORBNamingIOR",    ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
-	{ "ORBRootPOAIOR",   ORBIT_OPTION_STRING,  NULL }, /* FIXME: huh?          */
- 	{ "ORBIIOPIPName",   ORBIT_OPTION_STRING,  &orbit_ipname },
- 	{ "ORBIIOPIPSock",   ORBIT_OPTION_STRING,  &orbit_ipsock },
-	{ "ORBLocalOnly",    ORBIT_OPTION_BOOLEAN, &orbit_local_only },
+	{ "ORBid",              ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
+	{ "ORBImplRepoIOR",     ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
+	{ "ORBIfaceRepoIOR",    ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
+	{ "ORBNamingIOR",       ORBIT_OPTION_STRING,  NULL }, /* FIXME: unimplemented */
+	{ "ORBRootPOAIOR",      ORBIT_OPTION_STRING,  NULL }, /* FIXME: huh?          */
+ 	{ "ORBIIOPIPName",      ORBIT_OPTION_STRING,  &orbit_ipname },
+ 	{ "ORBIIOPIPSock",      ORBIT_OPTION_STRING,  &orbit_ipsock },
+	{ "ORBInitialMsgLimit", ORBIT_OPTION_INT,     &orbit_initial_recv_limit },
+	{ "ORBLocalOnly",       ORBIT_OPTION_BOOLEAN, &orbit_local_only },
 	/* warning: this option is a security risk unless used with LocalOnly */
-	{ "ORBIIOPIPv4",     ORBIT_OPTION_BOOLEAN, &orbit_use_ipv4 },
-	{ "ORBIIOPIPv6",     ORBIT_OPTION_BOOLEAN, &orbit_use_ipv6 },
-	{ "ORBIIOPUSock",    ORBIT_OPTION_BOOLEAN, &orbit_use_usocks },
-	{ "ORBIIOPUNIX",     ORBIT_OPTION_BOOLEAN, &orbit_use_usocks },
-	{ "ORBIIOPIrDA",     ORBIT_OPTION_BOOLEAN, &orbit_use_irda },
-	{ "ORBIIOPSSL",      ORBIT_OPTION_BOOLEAN, &orbit_use_ssl },
+	{ "ORBIIOPIPv4",        ORBIT_OPTION_BOOLEAN, &orbit_use_ipv4 },
+	{ "ORBIIOPIPv6",        ORBIT_OPTION_BOOLEAN, &orbit_use_ipv6 },
+	{ "ORBIIOPUSock",       ORBIT_OPTION_BOOLEAN, &orbit_use_usocks },
+	{ "ORBIIOPUNIX",        ORBIT_OPTION_BOOLEAN, &orbit_use_usocks },
+	{ "ORBIIOPIrDA",        ORBIT_OPTION_BOOLEAN, &orbit_use_irda },
+	{ "ORBIIOPSSL",         ORBIT_OPTION_BOOLEAN, &orbit_use_ssl },
 	/* warning: this option is a security risk */
-	{ "ORBHTTPIORs",     ORBIT_OPTION_BOOLEAN, &orbit_use_http_iors },
+	{ "ORBHTTPIORs",        ORBIT_OPTION_BOOLEAN, &orbit_use_http_iors },
 	/* warning: this option is a security risk */
-	{ "ORBSimpleUIDs",   ORBIT_OPTION_BOOLEAN, &orbit_use_genuid_simple },
-	{ "ORBDebugFlags",   ORBIT_OPTION_STRING,  &orbit_debug_options },
-	{ NULL,              0,                    NULL },
+	{ "ORBSimpleUIDs",      ORBIT_OPTION_BOOLEAN, &orbit_use_genuid_simple },
+	{ "ORBDebugFlags",      ORBIT_OPTION_STRING,  &orbit_debug_options },
+	{ NULL,                 0,                    NULL },
 };
