@@ -458,11 +458,13 @@ IOP_delete_profiles (CORBA_ORB orb,
 		return;
 
 	if (orb && *profiles == orb->profiles)
-		return;
-
-	g_slist_foreach (*profiles, (GFunc)IOP_profile_free, NULL);
-	g_slist_free (*profiles);
-	*profiles = NULL;
+		*profiles = NULL;
+	else {
+		g_slist_foreach (*profiles,
+				 (GFunc)IOP_profile_free, NULL);
+		g_slist_free (*profiles);
+		*profiles = NULL;
+	}
 }
 
 GSList *
@@ -610,8 +612,6 @@ IOP_generate_profiles (CORBA_Object obj)
 		obj->object_key = ORBit_OAObject_object_to_objkey (adaptor_obj);
 
 	obj->profile_list = orb->profiles;
-
-	ORBit_register_objref (obj);
 }
 
 static void
