@@ -9,10 +9,14 @@
 #include "../util/orbit-purify.h"
 #include "../orb-core/orbit-debug.h"
 #include "../GIOP/giop-debug.h"
+#include "../GIOP/giop-recv-buffer.h"
 #include "poa-macros.h"
-
+#include "poa-private.h"
 #include "orbit-poa.h"
 
+#define POA_LOCK(poa)   LINC_MUTEX_LOCK(poa->base.lock)
+#define POA_UNLOCK(poa) LINC_MUTEX_UNLOCK(poa->base.lock)
+
 static PortableServer_Servant ORBit_POA_ServantManager_use_servant(
 				     PortableServer_POA poa,
 				     ORBit_POAObject pobj,
@@ -94,13 +98,20 @@ static ORBit_POAObject
 ORBit_POACurrent_get_object (PortableServer_Current  obj,
 			     CORBA_Environment      *ev)
 {
+	ORBit_POAObject pobj = CORBA_OBJECT_NIL;
+
 	g_assert (obj && obj->parent.interface->type == ORBIT_ROT_POACURRENT);
 
-	poa_exception_val_if_fail (obj->orb->current_invocations != NULL,
-				   ex_PortableServer_Current_NoContext,
-				   CORBA_OBJECT_NIL);
+	LINC_MUTEX_LOCK (obj->orb->lock);
+	if (obj->orb->current_invocations == NULL)
+		CORBA_exception_set_system
+			(ev, ex_PortableServer_Current_NoContext,
+			 CORBA_COMPLETED_NO);
+	else
+		pobj = (ORBit_POAObject) obj->orb->current_invocations->data;
+	LINC_MUTEX_UNLOCK (obj->orb->lock);
 
-	return (ORBit_POAObject) obj->orb->current_invocations->data;
+	return pobj;
 }
 
 PortableServer_ClassInfo *
@@ -133,48 +144,62 @@ ORBit_classinfo_register (PortableServer_ClassInfo *ci)
 }
 
 static void
-check_child_poa_inuse (char               *name,
-		       PortableServer_POA  poa,
-		       gboolean           *is_inuse)
+check_child_poa_inuse_T (char               *name,
+			 PortableServer_POA  poa,
+			 gboolean           *is_inuse)
 {
 	if (ORBit_POA_is_inuse (poa, CORBA_TRUE, NULL))
 		*is_inuse = TRUE;
 }
 
 static void
-check_object_inuse (PortableServer_ObjectId *oid,
-		    ORBit_POAObject          pobj, 
-		    gboolean                *is_inuse)
+check_object_inuse_T (PortableServer_ObjectId *oid,
+		      ORBit_POAObject          pobj, 
+		      gboolean                *is_inuse)
 {
 	if (pobj->use_cnt > 0)
 		*is_inuse = TRUE;
 }
 
+static gboolean
+ORBit_POA_is_inuse_T (PortableServer_POA  poa,
+		      CORBA_boolean       consider_children,
+		      CORBA_Environment  *ev)
+{
+	gboolean is_inuse = FALSE;
+
+	if (poa->use_cnt > 0) 
+		return TRUE;
+
+	if (consider_children && poa->child_poas)
+		g_hash_table_foreach (poa->child_poas,
+				      (GHFunc) check_child_poa_inuse_T,
+				     &is_inuse);
+
+	if (!is_inuse && poa->oid_to_obj_map)
+		g_hash_table_foreach (poa->oid_to_obj_map,
+				      (GHFunc) check_object_inuse_T,
+				      &is_inuse);
+
+	return is_inuse;
+}
+
 gboolean
 ORBit_POA_is_inuse (PortableServer_POA  poa,
 		    CORBA_boolean       consider_children,
 		    CORBA_Environment  *ev)
 {
-	gboolean is_inuse = FALSE;
+	gboolean inuse;
 
-	if (poa->use_cnt > 0) 
-		return CORBA_TRUE;
+	POA_LOCK (poa);
+	inuse = ORBit_POA_is_inuse_T (poa, consider_children, ev);
+	POA_UNLOCK (poa);
 
-	if (consider_children && poa->child_poas)
-		g_hash_table_foreach (poa->child_poas,
-				      (GHFunc) check_child_poa_inuse,
-				     &is_inuse);
-
-	if (!is_inuse && poa->oid_to_obj_map)
-		g_hash_table_foreach (poa->oid_to_obj_map,
-				      (GHFunc) check_object_inuse,
-				      &is_inuse);
-
-	return is_inuse;
+	return inuse;
 }
 
-static PortableServer_ObjectId*
-ORBit_POA_new_system_objid (PortableServer_POA poa)
+static PortableServer_ObjectId *
+ORBit_POA_new_system_objid_T (PortableServer_POA poa)
 {
 	PortableServer_ObjectId *objid;
 
@@ -197,7 +222,7 @@ ORBit_POAObject_object_to_objkey (ORBit_POAObject pobj)
 {
 	ORBit_ObjectAdaptor  adaptor;
 	ORBit_ObjectKey     *objkey;
-	gpointer             mem;
+	guchar              *mem;
 
 	g_return_val_if_fail (pobj != NULL, NULL);
 
@@ -209,20 +234,22 @@ ORBit_POAObject_object_to_objkey (ORBit_POAObject pobj)
 	objkey->_buffer  = CORBA_sequence_CORBA_octet_allocbuf (objkey->_length);
 	objkey->_release = CORBA_TRUE;
 
-	mem = objkey->_buffer;
-	memcpy (mem, adaptor->adaptor_key._buffer, adaptor->adaptor_key._length);
-
+ 	mem = (guchar *) objkey->_buffer;
+ 	memcpy (mem, adaptor->adaptor_key._buffer, adaptor->adaptor_key._length);
+ 
 	mem += adaptor->adaptor_key._length;
-	memcpy (mem, pobj->object_id->_buffer, pobj->object_id->_length);
+ 	memcpy (mem, pobj->object_id->_buffer, pobj->object_id->_length);
 
 	return objkey;
 }
 
 static ORBit_POAObject
-ORBit_POA_object_id_lookup (PortableServer_POA             poa,
-			    const PortableServer_ObjectId *oid)
+ORBit_POA_object_id_lookup_T (PortableServer_POA             poa,
+			      const PortableServer_ObjectId *oid)
 {
-	return g_hash_table_lookup (poa->oid_to_obj_map, oid);
+	/* FIXME: needs some locking action */
+	return ORBit_RootObject_duplicate (
+		g_hash_table_lookup (poa->oid_to_obj_map, oid));
 }
 
 static void
@@ -284,6 +311,8 @@ ORBit_POA_destroy (PortableServer_POA  poa,
 	if (poa->life_flags & (ORBit_LifeF_Deactivating|ORBit_LifeF_Destroying))
 		return FALSE;	/* recursion */
 
+	/* FIXME: this needs some hard-core thread safety work ... */
+
 	poa->life_flags |= ORBit_LifeF_Destroying;
 
 	adaptors = poa->orb->adaptors;
@@ -325,7 +354,7 @@ ORBit_POA_destroy (PortableServer_POA  poa,
 	poa->life_flags &= ~ORBit_LifeF_Destroying;
 	ORBit_RootObject_release (poa);
 
-	return CORBA_TRUE;
+	return TRUE;
 }
 
 typedef struct TraverseInfo {
@@ -446,6 +475,9 @@ ORBit_POA_free_fn (ORBit_RootObject obj)
 	if (poa->name)
 		g_free (poa->name);
 
+	if (poa->base.lock)
+		g_mutex_free (poa->base.lock);
+
 	ORBit_RootObject_release_T (poa->orb);
 	ORBit_RootObject_release_T (poa->poa_manager);
 
@@ -536,7 +568,7 @@ ORBit_POA_set_policies (PortableServer_POA      poa,
 {
 	CORBA_unsigned_long i;
 
-	poa->p_thread              = PortableServer_ORB_CTRL_MODEL;
+	poa->p_thread              = PortableServer_SINGLE_THREAD_MODEL;
 	poa->p_lifespan            = PortableServer_TRANSIENT;
 	poa->p_id_uniqueness       = PortableServer_UNIQUE_ID;
 	poa->p_id_assignment       = PortableServer_SYSTEM_ID;
@@ -574,7 +606,7 @@ ORBit_POA_new (CORBA_ORB                  orb,
 	ORBit_RootObject_init ((ORBit_RootObject) poa, &ORBit_POA_epv);
 	/* released in ORBit_POA_destroy */
 	ORBit_RootObject_duplicate (poa);
-
+	
 	ORBit_POA_set_policies (poa, policies, ev);
 	if (ev->_major != CORBA_NO_EXCEPTION) {
 		ORBit_RootObject_release (poa);
@@ -603,6 +635,8 @@ ORBit_POA_new (CORBA_ORB                  orb,
 			(GHashFunc) ORBit_sequence_CORBA_octet_hash,
 			(GEqualFunc) ORBit_sequence_CORBA_octet_equal);
 
+	poa->base.lock = linc_mutex_new ();
+
 	ORBit_POAManager_register_poa (manager, poa);
 
 	return ORBit_RootObject_duplicate (poa);
@@ -629,11 +663,9 @@ ORBit_POA_obj_to_ref (PortableServer_POA  poa,
 	oid = pobj->object_id;
 
 	pobj->base.objref = ORBit_objref_new (poa->poa_manager->orb,
+					      &pobj->base,
 					      g_quark_from_string (type_id));
 
-	/* released by CORBA_Object_release */
-	pobj->base.objref->adaptor_obj = ORBit_RootObject_duplicate (pobj);
-
 	return ORBit_RootObject_duplicate (pobj->base.objref);
 }
 
@@ -718,9 +750,9 @@ ORBit_OAObject_Interface_type ORBit_POAObject_methods = {
  *  is allowed by the CORBA spec.
  */
 static ORBit_POAObject
-ORBit_POA_create_object (PortableServer_POA             poa,
-			 const PortableServer_ObjectId *objid,
-			 CORBA_Environment             *ev)
+ORBit_POA_create_object_T (PortableServer_POA             poa,
+			   const PortableServer_ObjectId *objid,
+			   CORBA_Environment             *ev)
 {
 	ORBit_POAObject newobj;
 
@@ -746,7 +778,7 @@ ORBit_POA_create_object (PortableServer_POA             poa,
 			memcpy (newobj->object_id->_buffer, objid->_buffer, objid->_length);
 		}
 		else
-			newobj->object_id = ORBit_POA_new_system_objid (poa);
+			newobj->object_id = ORBit_POA_new_system_objid_T (poa);
 	} else {
 		newobj->object_id           = PortableServer_ObjectId__alloc ();
 		newobj->object_id->_length  = objid->_length;
@@ -768,10 +800,10 @@ ORBit_POA_create_object (PortableServer_POA             poa,
  *  or NON_RETAIN.
  */
 static void
-ORBit_POA_activate_object (PortableServer_POA          poa, 
-			   ORBit_POAObject             pobj,
-			   PortableServer_ServantBase *servant, 
-			   CORBA_Environment          *ev) 
+ORBit_POA_activate_object_T (PortableServer_POA          poa, 
+			     ORBit_POAObject             pobj,
+			     PortableServer_ServantBase *servant, 
+			     CORBA_Environment          *ev) 
 {
 	PortableServer_ClassInfo *class_info;
 
@@ -963,13 +995,15 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 {
 	PortableServer_POA                   poa = pobj->poa;
 	PortableServer_ServantLocator_Cookie cookie;
-	PortableServer_ObjectId             *oid;
+	PortableServer_ObjectId             *oid = NULL;
 	PortableServer_ClassInfo            *klass;
 	ORBit_IMethod                       *m_data = NULL;
-	ORBitSkeleton                        skel = NULL;
 	ORBitSmallSkeleton                   small_skel = NULL;
 	gpointer                             imp = NULL;
 
+	if (poa)
+		POA_LOCK (poa);
+
 	if (!poa || !poa->poa_manager)
 		CORBA_exception_set_system (
 			ev, ex_CORBA_OBJECT_NOT_EXIST, 
@@ -984,24 +1018,26 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 						
 				poa->held_requests = g_slist_prepend (
 					poa->held_requests, recv_buffer);
-			}
-			else
+				
+				POA_UNLOCK (poa);
+				return;
+			} else
 				CORBA_exception_set_system (
 					ev, ex_CORBA_TRANSIENT,
 					CORBA_COMPLETED_NO);
-			return;
+			break;
 			
 		case PortableServer_POAManager_DISCARDING:
 			CORBA_exception_set_system (
 				ev, ex_CORBA_TRANSIENT,
 				CORBA_COMPLETED_NO);
-			return;
+			break;
 
 		case PortableServer_POAManager_INACTIVE:
 			CORBA_exception_set_system (
 				ev, ex_CORBA_OBJ_ADAPTER,
 				CORBA_COMPLETED_NO);
-			return;
+			break;
 
 		case PortableServer_POAManager_ACTIVE:
 			break;
@@ -1011,6 +1047,11 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 			break;
 		}
 	}
+	if (ev->_major != CORBA_NO_EXCEPTION) {
+		return_exception (recv_buffer, m_data, ev);
+		goto clean_out;
+	}
+
 	oid = pobj->object_id;
 
 	if (!pobj->servant) {
@@ -1023,7 +1064,7 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 			break;
 
 		case PortableServer_USE_DEFAULT_SERVANT:
-			ORBit_POA_activate_object (
+			ORBit_POA_activate_object_T (
 				poa, pobj, poa->default_servant, ev);
 			break;
 
@@ -1038,8 +1079,10 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 	}
 
 	pobj->use_cnt++;
+	LINC_MUTEX_LOCK (poa->orb->lock);
 	poa->orb->current_invocations =
 		g_slist_prepend (poa->orb->current_invocations, pobj);
+	LINC_MUTEX_UNLOCK (poa->orb->lock);
 	
 	if (ev->_major == CORBA_NO_EXCEPTION && !pobj->servant)
 		CORBA_exception_set_system (
@@ -1053,21 +1096,18 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 
 	klass = ORBIT_SERVANT_TO_CLASSINFO (pobj->servant);
 
-	if (klass->relay_call)
-		skel = klass->relay_call (
-			pobj->servant, recv_buffer, &imp);
-
-	if (!skel && klass->small_relay_call)
-		small_skel = klass->small_relay_call (
+	if (klass->impl_finder)
+		small_skel = klass->impl_finder (
 			pobj->servant, opname, (gpointer *)&m_data, &imp);
 
-	if (!skel && !small_skel)
+	/* FIXME: we can only do that in-proc (and for _is_a [and others?]) */
+	if (!small_skel)
 		small_skel = get_small_skel_CORBA_Object (
 			pobj->servant, opname,
 			(gpointer *)&m_data, &imp);
 	
-	if ((!small_skel && !skel) || !imp) {
-		if (!imp && (small_skel || skel)) {
+	if (!small_skel || !imp) {
+		if (!imp && small_skel) {
 			tprintf ("'%s' not implemented on %p",
 				 opname, pobj);
 			CORBA_exception_set_system (
@@ -1087,10 +1127,10 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 		goto clean_out;
 	}
 
-	if (skel)
-		skel (pobj->servant, recv_buffer, ev, imp);
+	ORBit_RootObject_duplicate (poa);
+	POA_UNLOCK (poa);
 
-	else if (recv_buffer) {
+	if (recv_buffer) {
 		struct ORBit_POA_invoke_data invoke_data;
 
 		invoke_data.small_skel = small_skel;
@@ -1102,6 +1142,9 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 	} else
 		small_skel (pobj->servant, ret, args, ctx, ev, imp);
 
+	POA_LOCK (poa);
+	ORBit_RootObject_release (poa);
+
  clean_out:
 	if (recv_buffer)
 		CORBA_exception_free (ev);
@@ -1121,26 +1164,135 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 			break;
 		}
 
+	LINC_MUTEX_LOCK (poa->orb->lock);
 	g_assert ((ORBit_POAObject)poa->orb->current_invocations->data == pobj);
 	poa->orb->current_invocations =
 		g_slist_remove (poa->orb->current_invocations, pobj);
+	LINC_MUTEX_UNLOCK (poa->orb->lock);
 	pobj->use_cnt--;
 
 	if (pobj->life_flags & ORBit_LifeF_NeedPostInvoke)
 		ORBit_POAObject_post_invoke (pobj);             
+
+	POA_UNLOCK (poa);
 }
 
 static ORBit_POAObject
-ORBit_POA_object_key_lookup (PortableServer_POA       poa,
-			     ORBit_ObjectKey         *objkey,
-			     PortableServer_ObjectId *object_id)
+ORBit_POA_object_key_lookup_T (PortableServer_POA       poa,
+			       ORBit_ObjectKey         *objkey,
+			       PortableServer_ObjectId *object_id)
 {
 	object_id->_buffer  = objkey->_buffer + ORBIT_ADAPTOR_PREFIX_LEN;
 	object_id->_length  = objkey->_length - ORBIT_ADAPTOR_PREFIX_LEN; 
 	object_id->_maximum = object_id->_length;
 	object_id->_release = CORBA_FALSE;
 
-	return ORBit_POA_object_id_lookup (poa, object_id);
+	return ORBit_POA_object_id_lookup_T (poa, object_id);
+}
+
+static void
+ORBit_POAObject_invoke_incoming_request (ORBit_POAObject    pobj,
+					 GIOPRecvBuffer    *recv_buffer,
+					 CORBA_Environment *opt_ev)
+{
+	CORBA_Environment real_ev, *ev;
+	
+	if (!opt_ev)
+		CORBA_exception_init ((ev = &real_ev));
+	else
+		ev = opt_ev;
+
+	if (ev->_major == CORBA_NO_EXCEPTION &&
+	    pobj != CORBA_OBJECT_NIL) {
+		CORBA_Identifier opname;
+
+		opname = giop_recv_buffer_get_opname (recv_buffer);
+		ORBit_POAObject_handle_request (pobj, opname, NULL, NULL, 
+						NULL, recv_buffer, ev);
+	}
+
+	ORBit_RootObject_release (pobj);
+
+	if (ev->_major != CORBA_NO_EXCEPTION) {
+		tprintf ("p %d, Method '%p' invoked with exception '%s'",
+			 getpid (), giop_recv_buffer_get_opname (recv_buffer), ev->_id);
+
+		return_exception (recv_buffer, NULL, ev);
+	}
+
+	if (!opt_ev)
+		CORBA_exception_free (ev);
+
+	giop_recv_buffer_unuse (recv_buffer);
+}
+
+static inline void
+push_request_T (GIOPThread       *thread,
+		ORBit_POAObject  *pobj,
+		GIOPRecvBuffer  **recv_buffer)
+{
+	giop_thread_request_push
+		(thread, (gpointer *)pobj, (gpointer *)recv_buffer);
+}
+
+typedef struct {
+	ORBit_POAObject  pobj;
+	GIOPRecvBuffer  *recv_buffer;
+} PoaIdleClosure;
+
+static gboolean
+poa_invoke_at_idle (gpointer data)
+{
+	PoaIdleClosure *pcl = data;
+
+	ORBit_POAObject_invoke_incoming_request
+		(pcl->pobj, pcl->recv_buffer, NULL);
+
+	g_free (data);
+
+	return FALSE;
+}
+
+static void
+push_request_idle (ORBit_POAObject  *pobj,
+		   GIOPRecvBuffer  **recv_buffer)
+{
+	PoaIdleClosure *pcl = g_new (PoaIdleClosure, 1);
+
+	pcl->pobj = *pobj;
+	pcl->recv_buffer = *recv_buffer;
+
+	/* FIXME: track pending idle impls
+	 * for cleanup at shutdown ... */
+	g_idle_add (poa_invoke_at_idle, pcl);
+	
+	*pobj = NULL;
+	*recv_buffer = NULL;
+}
+
+static gboolean
+poa_recv_is_oneway (ORBit_POAObject pobj,
+		    GIOPRecvBuffer *recv_buffer)
+{
+	ORBit_IMethod            *m_data = NULL;
+	gpointer                  imp = NULL;
+	PortableServer_ClassInfo *klass;
+
+	g_return_val_if_fail (pobj != CORBA_OBJECT_NIL, FALSE);
+
+	klass = ORBIT_SERVANT_TO_CLASSINFO (pobj->servant);
+
+	if (!klass->impl_finder)
+		return FALSE;
+
+	klass->impl_finder ( pobj->servant,
+			     giop_recv_buffer_get_opname (recv_buffer),
+			     (gpointer *)&m_data, &imp);
+
+	if (m_data && m_data->flags & ORBit_I_METHOD_1_WAY)
+		return TRUE;
+
+	return FALSE;
 }
 
 static void
@@ -1148,14 +1300,15 @@ ORBit_POA_handle_request (PortableServer_POA poa,
 			  GIOPRecvBuffer    *recv_buffer,
 			  ORBit_ObjectKey   *objkey)
 {
-	PortableServer_ObjectId object_id;
 	ORBit_POAObject         pobj;
-	CORBA_Identifier        opname;
 	CORBA_Environment       env;
+	PortableServer_ObjectId object_id;
 
 	CORBA_exception_init (&env);
 
-	pobj = ORBit_POA_object_key_lookup (poa, objkey, &object_id);
+	POA_LOCK (poa);
+
+	pobj = ORBit_POA_object_key_lookup_T (poa, objkey, &object_id);
 
 	if (!pobj)
 		switch (poa->p_request_processing) {
@@ -1168,34 +1321,61 @@ ORBit_POA_handle_request (PortableServer_POA poa,
 
 		case PortableServer_USE_DEFAULT_SERVANT: /* drop through */
 		case PortableServer_USE_SERVANT_MANAGER:
-			pobj = ORBit_POA_create_object (poa, &object_id, &env);
+			pobj = ORBit_POA_create_object_T (poa, &object_id, &env);
 			break;
 
 		default:
 			g_assert_not_reached ();
 			break;
 		}
-
+	
 	if (!pobj)
 		CORBA_exception_set_system (
 			&env, ex_CORBA_OBJECT_NOT_EXIST, 
 			CORBA_COMPLETED_NO);
 	else {
-		opname = giop_recv_buffer_get_opname (recv_buffer);
-		
-		ORBit_POAObject_handle_request (pobj, opname, NULL, NULL, 
-						NULL, recv_buffer, &env);
+		switch (poa->p_thread) {
+			case PortableServer_SINGLE_THREAD_MODEL:
+				if (giop_threaded ())
+					push_request_T (giop_thread_get_main (),
+							&pobj, &recv_buffer);
+				break;
+			case PortableServer_ORB_CTRL_MODEL: {
+				ORBit_ObjectAdaptor adaptor = (ORBit_ObjectAdaptor) poa;
+				
+				switch (adaptor->thread_hint) {
+				case ORBIT_THREAD_HINT_ONEWAY_AT_IDLE:
+					if (!poa_recv_is_oneway (pobj, recv_buffer))
+						push_request_T (giop_thread_get_main (),
+								&pobj, &recv_buffer);
+					/* drop through */
+				case ORBIT_THREAD_HINT_ALL_AT_IDLE:
+					push_request_idle (&pobj, &recv_buffer);
+					break;
+				case ORBIT_THREAD_HINT_NONE:
+					if (giop_threaded ())
+						push_request_T (giop_thread_get_main (),
+								&pobj, &recv_buffer);
+					break;
+				default:
+					g_warning ("Binning incoming requests in threaded mode");
+					giop_recv_buffer_unuse (recv_buffer);
+					recv_buffer = NULL;
+					pobj = NULL;
+					break;
+				}
+				break;
+			}
+		default:
+			g_assert_not_reached ();
+			break;
+		}
 	}
-
+	
  send_sys_ex:
-	if (env._major != CORBA_NO_EXCEPTION) {
-		tprintf ("p %d, Method '%p' invoked with exception '%s'",
-			 getpid (), giop_recv_buffer_get_opname (recv_buffer), env._id);
+	ORBit_POAObject_invoke_incoming_request (pobj, recv_buffer, &env);
 
-		return_exception (recv_buffer, NULL, &env);
-	}
-
-	CORBA_exception_free (&env);
+	POA_UNLOCK (poa);
 }
 
 static PortableServer_Servant
@@ -1341,7 +1521,7 @@ PortableServer_ObjectId_to_string (PortableServer_ObjectId *id,
 	poa_sys_exception_val_if_fail (memchr (id->_buffer, '\0', id->_length),
 				       ex_CORBA_BAD_PARAM, NULL);
 
-	str = CORBA_string_alloc (id->_length);
+	str = CORBA_string_alloc (id->_length + 1);
 	memcpy (str, id->_buffer, id->_length);
 	str [id->_length] = '\0';
 
@@ -1359,7 +1539,7 @@ PortableServer_ObjectId_to_wstring (PortableServer_ObjectId *id,
 	poa_sys_exception_val_if_fail (memchr (id->_buffer, '\0', id->_length),
 				       ex_CORBA_BAD_PARAM, NULL);
   
-	retval = CORBA_wstring_alloc (id->_length);
+	retval = CORBA_wstring_alloc (id->_length + 1);
 	for (i = 0; i < id->_length; i++)
 		retval [i] = id->_buffer [i];
 	retval [id->_length] = '\0';
@@ -1647,22 +1827,29 @@ PortableServer_POA_activate_object (PortableServer_POA            poa,
 				    const PortableServer_Servant  p_servant,
 				    CORBA_Environment            *ev)
 {
+	PortableServer_ObjectId    *result;
 	PortableServer_ServantBase *servant = p_servant;
 	ORBit_POAObject             newobj;
 
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 	poa_sys_exception_val_if_fail (servant != NULL, ex_CORBA_BAD_PARAM, NULL);
 
+	POA_LOCK (poa);
+
+	/* FIXME: unlock on preconditions */
 	poa_exception_val_if_fail (IS_RETAIN (poa), ex_PortableServer_POA_WrongPolicy, NULL);
 	poa_exception_val_if_fail (IS_SYSTEM_ID (poa), ex_PortableServer_POA_WrongPolicy, NULL);
 	poa_exception_val_if_fail (IS_MULTIPLE_ID (poa) || (IS_UNIQUE_ID (poa) && servant->_private == NULL),
 				   ex_PortableServer_POA_ServantAlreadyActive, NULL);
 
-	newobj = ORBit_POA_create_object (poa, NULL, ev);
-	ORBit_POA_activate_object (poa, newobj, servant, ev);
+	newobj = ORBit_POA_create_object_T (poa, NULL, ev);
+	ORBit_POA_activate_object_T (poa, newobj, servant, ev);
 
-	return (PortableServer_ObjectId *)
-			ORBit_sequence_CORBA_octet_dup (newobj->object_id);
+	result = ORBit_sequence_CORBA_octet_dup (newobj->object_id);
+
+	POA_UNLOCK (poa);
+
+	return result;
 }
 
 void
@@ -1671,35 +1858,42 @@ PortableServer_POA_activate_object_with_id (PortableServer_POA             poa,
 					    const PortableServer_Servant   p_servant,
 					    CORBA_Environment             *ev)
 {
+	ORBit_POAObject pobj;
 	PortableServer_ServantBase *servant = p_servant;
-	ORBit_POAObject             newobj;
 
 	poa_sys_exception_if_fail (poa != NULL, ex_CORBA_INV_OBJREF);
 	poa_sys_exception_if_fail (objid != NULL, ex_CORBA_BAD_PARAM);
 	poa_sys_exception_if_fail (servant != NULL, ex_CORBA_BAD_PARAM);
 
+	POA_LOCK (poa);
+
 	poa_exception_if_fail (IS_RETAIN (poa), ex_PortableServer_POA_WrongPolicy);
 
-	newobj = ORBit_POA_object_id_lookup (poa, objid);
+	pobj = ORBit_POA_object_id_lookup_T (poa, objid);
 
-	if (newobj && newobj->servant) {
+	if (pobj && pobj->servant)
 		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
 				     ex_PortableServer_POA_ObjectAlreadyActive, 
 				     NULL);
-		return;
-	}
 
-	if (IS_UNIQUE_ID (poa) && servant->_private != NULL) {
+	else if (IS_UNIQUE_ID (poa) && servant->_private != NULL)
 		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
 				     ex_PortableServer_POA_ServantAlreadyActive,
 				     NULL);
-		return;
+	else {
+		ORBit_POAObject newobj;
+
+		if (!pobj)
+			newobj = ORBit_POA_create_object_T (poa, objid, ev);
+		else
+			newobj = pobj;
+
+		ORBit_POA_activate_object_T (poa, newobj, servant, ev);
 	}
 
-	if (!newobj)
-		newobj = ORBit_POA_create_object(poa, objid, ev);
+	ORBit_RootObject_release (pobj);
 
-	ORBit_POA_activate_object (poa, newobj, servant, ev);
+	POA_UNLOCK (poa);
 }
 
 void
@@ -1712,12 +1906,18 @@ PortableServer_POA_deactivate_object (PortableServer_POA             poa,
 	poa_sys_exception_if_fail (poa != NULL, ex_CORBA_INV_OBJREF);
 	poa_sys_exception_if_fail (oid != NULL, ex_CORBA_BAD_PARAM);
 
+	POA_LOCK (poa);
+
 	poa_exception_if_fail (IS_RETAIN (poa), ex_PortableServer_POA_WrongPolicy);
 
-	pobj = ORBit_POA_object_id_lookup (poa, oid);
+	pobj = ORBit_POA_object_id_lookup_T (poa, oid);
 
 	if (pobj && pobj->servant)
 		ORBit_POA_deactivate_object (poa, pobj, CORBA_TRUE, CORBA_FALSE);
+
+	POA_UNLOCK (poa);
+
+	ORBit_RootObject_release (pobj);
 }
 
 CORBA_Object
@@ -1725,15 +1925,22 @@ PortableServer_POA_create_reference (PortableServer_POA  poa,
 				     const CORBA_char   *intf,
 				     CORBA_Environment  *ev)
 {
+	CORBA_Object obj;
 	ORBit_POAObject pobj;
 
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 
+	POA_LOCK (poa);
+
 	poa_exception_val_if_fail (IS_SYSTEM_ID (poa), ex_PortableServer_POA_WrongPolicy, NULL);
 
-	pobj = ORBit_POA_create_object (poa, NULL, ev);
+	pobj = ORBit_POA_create_object_T (poa, NULL, ev);
 
-	return ORBit_POA_obj_to_ref (poa, pobj, intf, ev);
+	obj = ORBit_POA_obj_to_ref (poa, pobj, intf, ev);
+
+	POA_UNLOCK (poa);
+
+	return obj;
 }
 
 CORBA_Object
@@ -1742,16 +1949,28 @@ PortableServer_POA_create_reference_with_id (PortableServer_POA             poa,
 					     const CORBA_char              *intf,
 					     CORBA_Environment             *ev)
 {
-	ORBit_POAObject	pobj;
+	CORBA_Object obj;
+	ORBit_POAObject	pobj, newobj;
 
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 	poa_sys_exception_val_if_fail (oid != NULL, ex_CORBA_BAD_PARAM, NULL);
 
-	pobj = ORBit_POA_object_id_lookup (poa, oid);
-	if (!pobj)
-		pobj = ORBit_POA_create_object (poa, oid, ev);
+	POA_LOCK (poa);
 
-	return ORBit_POA_obj_to_ref (poa, pobj, intf, ev);
+	pobj = ORBit_POA_object_id_lookup_T (poa, oid);
+	if (!pobj)
+		newobj = ORBit_POA_create_object_T (poa, oid, ev);
+	else
+		newobj = CORBA_OBJECT_NIL;
+
+	obj = ORBit_POA_obj_to_ref (poa, pobj, intf, ev);
+
+	if (!newobj)
+		ORBit_RootObject_release (pobj);
+
+	POA_UNLOCK (poa);
+
+	return obj;
 }
 
 PortableServer_ObjectId *
@@ -1760,6 +1979,7 @@ PortableServer_POA_servant_to_id (PortableServer_POA            poa,
 				  CORBA_Environment            *ev)
 {
 	PortableServer_ServantBase *servant = p_servant;
+	PortableServer_ObjectId    *objid;
 	ORBit_POAObject             pobj = servant->_private;
 	gboolean                    defserv = IS_USE_DEFAULT_SERVANT (poa);
 	gboolean                    retain = IS_RETAIN (poa);
@@ -1769,19 +1989,23 @@ PortableServer_POA_servant_to_id (PortableServer_POA            poa,
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 	poa_sys_exception_val_if_fail (servant != NULL, ex_CORBA_BAD_PARAM, NULL);
 
+	POA_LOCK (poa);
+
 	poa_exception_val_if_fail (defserv || (retain && (unique || implicit)),
 				   ex_PortableServer_POA_WrongPolicy, NULL);
 
+	objid = NULL;
 	if (retain && unique && pobj && pobj->servant == servant)
-		return ORBit_sequence_CORBA_octet_dup (pobj->object_id);
+		objid = ORBit_sequence_CORBA_octet_dup (pobj->object_id);
 
 	else if (retain && implicit && (!unique || !pobj)) {
-		pobj = ORBit_POA_create_object (poa, NULL, ev);
-		ORBit_POA_activate_object (poa, pobj, servant, ev);
+		pobj = ORBit_POA_create_object_T (poa, NULL, ev);
+		ORBit_POA_activate_object_T (poa, pobj, servant, ev);
 
-		return ORBit_sequence_CORBA_octet_dup (pobj->object_id);
-	}
-	else {
+		objid = ORBit_sequence_CORBA_octet_dup (pobj->object_id);
+
+	} else {
+		GSList *l;
 		/*
 		 * FIXME:
 		 * This handles case 3 of the spec; but is broader:
@@ -1790,21 +2014,26 @@ PortableServer_POA_servant_to_id (PortableServer_POA            poa,
 		 * The stricter form could be implemented, 
 		 * but it would only add more code...
 		 */
-		GSList *l = poa->orb->current_invocations;
-
-		for ( ; l; l = l->next) {
+		LINC_MUTEX_LOCK (poa->orb->lock);
+		for (l = poa->orb->current_invocations; l; l = l->next) {
 			ORBit_POAObject pobj = l->data;
 
 			if (pobj->servant == servant)
-				return ORBit_sequence_CORBA_octet_dup (pobj->object_id);
+				objid = ORBit_sequence_CORBA_octet_dup (pobj->object_id);
 		}
+		LINC_MUTEX_UNLOCK (poa->orb->lock);
 	}
 
-	CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
-			     ex_PortableServer_POA_ServantNotActive,
-			     NULL);
+	if (!objid)
+		CORBA_exception_set
+			(ev, CORBA_USER_EXCEPTION,
+			 ex_PortableServer_POA_ServantNotActive,
+			 NULL);
 
-	return NULL;
+
+	POA_UNLOCK (poa);
+
+	return objid;
 }
 
 CORBA_Object
@@ -1814,6 +2043,7 @@ PortableServer_POA_servant_to_reference (PortableServer_POA            poa,
 {
 	PortableServer_ServantBase *servant = p_servant;
 	ORBit_POAObject             pobj = servant->_private;
+	CORBA_Object                result;
 	gboolean                    retain = IS_RETAIN (poa);
 	gboolean                    implicit = IS_IMPLICIT_ACTIVATION (poa);
 	gboolean                    unique = IS_UNIQUE_ID (poa);
@@ -1821,24 +2051,25 @@ PortableServer_POA_servant_to_reference (PortableServer_POA            poa,
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, CORBA_OBJECT_NIL);
 	poa_sys_exception_val_if_fail (servant != NULL, ex_CORBA_BAD_PARAM, CORBA_OBJECT_NIL);
 
+	POA_LOCK (poa);
 	poa_exception_val_if_fail (retain && (unique || implicit),
 				   ex_PortableServer_POA_WrongPolicy,
 				   CORBA_OBJECT_NIL);
 
+	result = CORBA_OBJECT_NIL;
 	if (retain && unique && pobj)
 		if (pobj->base.objref)
-			return ORBit_RootObject_duplicate (
-				pobj->base.objref);
+			result = ORBit_RootObject_duplicate (pobj->base.objref);
 		else
-			return ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
+			result = ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
 
 	else if (retain && implicit && (!unique || !pobj)) {
-		pobj = ORBit_POA_create_object (poa, NULL, ev);
-		ORBit_POA_activate_object (poa, pobj, servant, ev);
+		pobj = ORBit_POA_create_object_T (poa, NULL, ev);
+		ORBit_POA_activate_object_T (poa, pobj, servant, ev);
 
-		return ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
-	}
-	else {
+		result = ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
+	} else {
+		GSList *l;
 		/*
 		 * FIXME:
 		 * This case deals with "invoked in the context of
@@ -1848,21 +2079,25 @@ PortableServer_POA_servant_to_reference (PortableServer_POA            poa,
 		 * go backward from servant to pobj to use_cnt, but we
 		 * dont do this since forward search is more general 
 		 */
-		GSList *l = poa->orb->current_invocations;
-
-		for ( ; l; l = l->next) {
+		LINC_MUTEX_LOCK (poa->orb->lock);
+		for (l = poa->orb->current_invocations; l; l = l->next) {
 			ORBit_POAObject pobj = l->data;
 
 			if (pobj->servant == servant)
-				return ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
+				result = ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
 		}
+		LINC_MUTEX_UNLOCK (poa->orb->lock);
 	}
 
-	CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
-			     ex_PortableServer_POA_ServantNotActive,
-			     NULL);
+	if (result == CORBA_OBJECT_NIL)
+		CORBA_exception_set
+			(ev, CORBA_USER_EXCEPTION,
+			 ex_PortableServer_POA_ServantNotActive,
+			 NULL);
 
-	return NULL;
+	POA_UNLOCK (poa);
+
+	return result;
 }
 
 PortableServer_Servant
@@ -1924,24 +2159,34 @@ PortableServer_POA_id_to_servant (PortableServer_POA             poa,
 				  const PortableServer_ObjectId *object_id,
 				  CORBA_Environment             *ev)
 {
+	PortableServer_Servant servant = NULL;
+
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 	poa_sys_exception_val_if_fail (object_id != NULL, ex_CORBA_BAD_PARAM, NULL);
 
+	POA_LOCK (poa);
+
 	poa_exception_val_if_fail (IS_USE_DEFAULT_SERVANT (poa) || IS_RETAIN (poa),
 				   ex_PortableServer_POA_WrongPolicy, NULL);
 
 	if (IS_RETAIN (poa)) {
-		ORBit_POAObject pobj = ORBit_POA_object_id_lookup (poa, object_id);
+		ORBit_POAObject pobj = ORBit_POA_object_id_lookup_T (poa, object_id);
 
 		if (pobj && pobj->servant)
-			return pobj->servant;
+			servant = pobj->servant;
+
+		ORBit_RootObject_release (pobj);
 
 	} else if (poa->default_servant)
-		return poa->default_servant;
+		servant = poa->default_servant;
 
-	CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
-			     ex_PortableServer_POA_ObjectNotActive,
-			     NULL);
+	if (!servant)
+		CORBA_exception_set
+			(ev, CORBA_USER_EXCEPTION,
+			 ex_PortableServer_POA_ObjectNotActive,
+			 NULL);
+
+	POA_UNLOCK (poa);
 
 	return NULL;
 }
@@ -1951,22 +2196,37 @@ PortableServer_POA_id_to_reference (PortableServer_POA             poa,
 				    const PortableServer_ObjectId *object_id,
 				    CORBA_Environment             *ev)
 {
+	CORBA_Object obj;
 	ORBit_POAObject pobj;
 
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 	poa_sys_exception_val_if_fail (object_id != NULL, ex_CORBA_BAD_PARAM, NULL);
 
+	POA_LOCK (poa);
+
 	poa_exception_val_if_fail (IS_RETAIN (poa), ex_PortableServer_POA_WrongPolicy, NULL);
 
-	pobj = ORBit_POA_object_id_lookup (poa, object_id);
+	pobj = ORBit_POA_object_id_lookup_T (poa, object_id);
 	if (!pobj || !pobj->servant) {
 		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
 				     ex_PortableServer_POA_ObjectNotActive, NULL);
-		return CORBA_OBJECT_NIL;
+		obj = CORBA_OBJECT_NIL;
+	} else {
+		if (pobj->base.objref)
+			obj = ORBit_RootObject_duplicate (pobj->base.objref);
+		else
+			obj = ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
 	}
 
-	if (pobj->base.objref)
-		return ORBit_RootObject_duplicate (pobj->base.objref);
+	POA_UNLOCK (poa);
+	ORBit_RootObject_release (pobj);
+	
+	return obj;
+}
 
-	return ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
+void
+ORBit_poa_init (void)
+{
+	_ORBit_poa_manager_lock = linc_mutex_new ();
+	giop_thread_set_main_handler (ORBit_POAObject_invoke_incoming_request);
 }
